// +---------------------------------------------------------------------------+
//
//      Program:    hott_gps_sensor.c
//
//      Purpose:    HoTT GPS sensor for model planes: 
//					Gets GPS data information from a GPS that
//					is connected to the HW UART. Needed GPS telegrams are
//					$GPRMC and $GPGGA. The data is converted and sent via
//					the HoTT telemetrie interface to a ground station.
//      
//      Target:     ATmega8a
//
//      Author:     Herbert Dingfelder, DL5NEG
//
//      License:    GPL 3
//
//      History:    14.02.2016  - Initial Coding, based on hott_gps_prestudy.c
//								- Adaptatin from ATtiny13 to Atmega8 started
//								- UART timer changed to 16bit counter, since in the
//								  mega8, only the 16bit timer has two compare units
//                  15.02.2016  - Hardware UART activated for receiving data from GPS
//					16.02.2016	- First few bits of information extracted successfully
//								  from the GPS data stream
//					17.02.2016	- Interupt implemented to read GPS data from the HW
//								  UART simultaniously with the handling of the SW UART
//								  that does the HoTT interface. Concept urned out to be 
//								  highly time critical to avoid negative effects on HoTT
//								  SW-UART. Therefore the ISR was strongly optimized to 
//								  stay within 10-12µs at Fc=3.686MHz.
//					18.02.2016 	- Readout of float values works, the implementation is now
//								  robust towards different GPS receivers that deliver
//								  different numbers of digits for the data fields
//					19.02.2016 	- All values that transmitter shows are now extracted and 
//								  sent via the HoTT interface
//					21.02.2016	- Code cleanup for first beta version
//								  Crystal changed to 11.059MHz (higher fc give better precision
//								  on HoTT interface and allows for more instructions during ISR
//								  Tests with two different GPS receivers successfull on lab bench
//								  => Version 0.9 released
//					28.02.2016	- Additional tests done with the sensor from a car were also
//								  successfull, no complications during data aquistion or in
//								  the post processing afterwards.
//					29.02.2016	- Added the parallel behaviour as general HoTT module, to
//								  report other information like battery voltage.
//								- Battery voltage reading and reporting tested, works great
//								  in parallel to the GPS recording.
//								- Significant code cleanup after adding the voltage stuff
//								  => Version 1.0 released
//
// +---------------------------------------------------------------------------+

#define F_CPU 11059000
#include <util/delay.h>
#include <avr/io.h>
#include <stdio.h>
#include <string.h>
#include <avr/interrupt.h>
#include <avr/wdt.h>
#include <avr/sleep.h>
#include <avr/cpufunc.h>
#include <stdlib.h>


#define HOTT_SENSOR_ADR_BYTE1		0x80 	// the HoTT master starts a requests always with this byte
#define HOTT_GEN_MODULE_SIG			0x8d 	// for second byte of a request: signature is 0x8d for the general module
#define HOTT_GPS_MODULE_SIG			0x8a 	// for second byte of a request: signature is 0x8a for the GPS module
#define HOTT_START_BYTE				0x7c 	// first byte in response telegram is always 0x7c
#define HOTT_GPS_TEXT_ID			0xa0	// text id of the GPS sensor

#define LOW_CELL_VOLTAGE_THRESHOLD	170		// low single cell voltage warning level in 20mV, we choose 3.4V (3.4V/20mV=170)
#define LOW_FUEL_BAR_VALUE			(3400*3)// fuel bar shall read 0% when voltages had droped to 3.4V per cell	
#define MAX_FUEL_BAR_VALUE			(4200*3)// fuel bar shall read 100% when voltages are as high as 4.2V per cell

#define LIPO_TAP3_ADC_CHANNEL 0     // LiPo battery tap 3 (overall voltage) is connected via divider to ADC 0          

#define HOTT_PORT PORTB				// HoTT interface is connected to PB0
#define HOTT_DDR  DDRB
#define HOTT_IN   PINB
#define HOTT_BIT  _BV(0)

#define TXD_PORT HOTT_PORT          // TXD and RXD are both connected to the one HoTT pin
#define TXD_PIN  HOTT_BIT
#define TXD_DDR  HOTT_DDR
#define RXD_PIN  HOTT_BIT
#define RXD_READ HOTT_IN

#define LED_PORT PORTB				// LED is connected to PB1 (for debugging only, not needed for normal operation)
#define LED_DDR  DDRB
#define LED_BIT  _BV(1)


// ----- macros to switch TXD to low/high, to enhance code readability -----

#define UART_TXD_LOW    (TXD_PORT &= ~(TXD_PIN) )       // macro to set TXD low
#define UART_TXD_HIGH   (TXD_PORT |=  (TXD_PIN) )       // macro to set TXD high
#define UART_OPMODE_TX	(TXD_DDR |=   (TXD_PIN) ) 		// macro to switch UART pin to output
#define UART_OPMODE_RX	(TXD_DDR &=  ~(TXD_PIN) ) 		// macro to switch UART pin to input
#define UART_RXD        (RXD_READ & (RXD_PIN) )         // macro to read RXD back
#define LED_ON          (LED_PORT |=  LED_BIT)          // macro to switch on LED
#define LED_OFF         (LED_PORT &= ~LED_BIT)          // macro to switch off LED


void reset_handler(void);
void send_byte_via_sw_uart(char byte_to_send);
int receive_byte_via_sw_uart(void);
void wait_for_next_uart_cycle(void);
void init_RS232_timer(void);
uint8_t MSB(uint16_t word_var);
uint8_t LSB(uint16_t word_var);
void read_cell_voltages(void);


// For reasons of simplicity and memory efficiency, we keep as GPS data and voltage data variables global

uint8_t  gps_alarm_beep = 0x00;                     // default for alarm beep is off
uint8_t  gps_alarm_bitmask = 0x00;					// default for alarm bit mask is no alarm
uint8_t  gps_alarm_no_fix = 0x00;					// default for gps fix alarm is no alarm
uint8_t  gps_flightdir_2deg = 0;					// default for flight direction is 0° (2° steps)
uint16_t gps_speed_kmh =0;							// default for speed is 0km/h
uint8_t  gps_north_south = 0;						// default for north/south is north (0)
uint16_t gps_lat_ddmm = 0;							// default for latitude is 0°0'
uint16_t gps_lat_dec_mmmm = 0;						// default for sub decimal minutes is 0
uint8_t  gps_east_west = 0;							// default for east/west is east (0)
uint16_t gps_long_ddmm = 0;							// default for longitude is 0°0'
uint16_t gps_long_dec_mmmm = 0;						// default for sub decimal minutes is 0
uint16_t gps_distance = 0;							// default for distance is 0m
uint16_t gps_altitude = 500+0;						// default for altitude is 0m (with 500 offset per definition)
uint16_t gps_climb_mps = 30000+0;					// default for climb in m/s is 0m/s (in 0.01m/s with 30000 offset)
uint8_t  gps_climb_mp3s = 120+0;					// default for climb in m/3s is 0m/s (with 120 offset)
uint8_t  gps_no_sats = 0;							// default for number of satelites is 0
uint8_t  gps_fix_mode = 0;							// default for fix mode is kept to 0, interpretation not clear
uint8_t  gps_dir_mod_home = 0;						// default for direction from model to home is 0° (2° steps)
uint8_t  gps_time_hour = 0;                         // default for the GPS time is 00:00:00
uint8_t  gps_time_min = 0;                          // default for the GPS time is 00:00:00
uint8_t  gps_time_sec =0;                           // default for the GPS time is 00:00:00

uint8_t battery_voltage_in_100mV=124;               // default for battery voltage (just some value above warning threshold)
uint16_t battery_voltage_in_mV=12400;				// default for battery voltage in mV (shown in fuel ml field) is 12400mV
uint8_t cell1_voltage=200;                          // default for cell voltage within a normal range (200x20mV=4.0V)
uint8_t cell2_voltage=200;                          // default for cell voltage within a normal range (200x20mV=4.0V)
uint8_t cell3_voltage=200;                          // default for cell voltage within a normal range (200x20mV=4.0V)
uint8_t temperature1=20;							// default for temperature1 is 0°C (20 offset per protocol definiton)
uint8_t temperature2=20;							// default for temperature2 is 0°C (20 offset per protocol definiton)
uint8_t fuel_bar=100;                         		// default for fuel bar graph is 100%
uint8_t alarm_beep = 0x00;                          // default for alarm beep is off

uint16_t checksum;	// global variable checksum (helps so safe memory to have this global)

#define BUFFER_SIZE 100
uint8_t gps_buffer_gprmc[BUFFER_SIZE];
uint8_t gps_buffer_gpgga[BUFFER_SIZE];
uint8_t temp_buffer[BUFFER_SIZE];
uint8_t second_tick=0;

// ============================= main program starts here ======================
int main(void)
{
  
    uint8_t received_databyte;
    
    char *substring_ptr;
    uint16_t old_altitude=0;
	int16_t diff_altitude;
    double speed_knots_f;
    double course_f; 
    double lat_f;
    double lon_f;
    double altitude_f;
    
    
    
    // ----- here we go -----
    
    reset_handler();                                    // do all the regular port and register settings
                                                        //   and start the needed UART counter
   
	for(;;)                                             // --- main loop starts here ---
	{

		received_databyte = receive_byte_via_sw_uart();	// wait for next character on HoTT interface
		
		if(received_databyte != HOTT_SENSOR_ADR_BYTE1 )	// if it is not the first byte of a valid address
			continue;									//   start a new loop immediately

		received_databyte = receive_byte_via_sw_uart();	// wait for next character on Hott interface
		
		if( received_databyte == HOTT_GPS_MODULE_SIG )	// if 2nd byte of the request is that of the GPS sensor...  
		{
		
			// --- now we know we have been called and shall respond with our information, lets put that together ---
				
			// In order to make sure that we always process consistend data, we copy the needed data buffer
			// into a temporary buffer first. During that copy, we lock the interupts, to avoid change of the
			// data while we copy. The copy process takes 192µs at Fc=3.686MHz, which is well below the problematic
			// time when the GPS UART would start to lose incomming data (about 1ms).
			cli();
			memcpy(temp_buffer, gps_buffer_gprmc, BUFFER_SIZE);
			sei();
		
			//example messages from different GPS receiver as test vectors, to secure it works with all of them
			//strcpy(temp_buffer, "$GPRMC,225446,A,4916.987654,N,12311.12,W,000.5,054.7,191194,020.3,E*68");
			//strcpy(temp_buffer, "$GPRMC,225446,A,4916.1234,N,12311.1234,W,000.5,054.7,191194,020.3,E*68");
			//strcpy(temp_buffer, "$GPRMC,220516,A,5133.82,N,00042.24,W,173.8,231.8,130694,004.2,W*70");
			//strcpy(temp_buffer, "$GPRMC,120134.000,A,4922.1269,N,01104.7201,E,0.32,154.10,180715,,,A*60");
			//strcpy(temp_buffer, "$GPRMC,180411.00,A,4922.18235,N,01104.81545,E,0.080,,160216,,,A*7E");
			
			substring_ptr = memchr(temp_buffer, ',', BUFFER_SIZE);			// get a pointer to the first position of a comma in the message

			gps_time_hour = (substring_ptr[1]-48)*10 + (substring_ptr[2]-48);	// extract the time, which is typically always available from 
			gps_time_min  = (substring_ptr[3]-48)*10 + (substring_ptr[4]-48);	// the receiver (exception is a cold start)
			gps_time_sec  = (substring_ptr[5]-48)*10 + (substring_ptr[6]-48);
			
			substring_ptr = memchr(substring_ptr+1, ',', BUFFER_SIZE);		// get a pointer to the next position of a comma in the message
			
			if( substring_ptr[1]=='A' )										// read out if the GPS receiver has a valid fix
				gps_alarm_no_fix=0;											// reset alarm bit if fix is ok       
			else
				gps_alarm_no_fix=1;											// set alarm bit if GPS receiver does not have a valid signal
			
			
			if(gps_alarm_no_fix==0)											// only if GPS receiver has a valid signal, send postion and other data
			{
				
				substring_ptr = memchr(substring_ptr+1, ',', BUFFER_SIZE);		// get a pointer to the next position of a comma in the message
				
				lat_f = atof(substring_ptr+1);									// read the latitude as floating point value from the message
																				// (float looses some very slight precision, but is much more robust
																				//  against different numbers of digits in the GPS receiver implementation,
																				//  so this way should work with basically all kind of GPS receivers.
				
				gps_lat_ddmm = (int)lat_f;										// extract the integer part of the latitude
				gps_lat_dec_mmmm = (lat_f-gps_lat_ddmm)*10000;					// extract the post-decimal part of the latitude as 4digit number as wanted by HoTT
		
				substring_ptr = memchr(substring_ptr+1, ',', BUFFER_SIZE);		// get a pointer to the next position of a comma in the message
				
				if( substring_ptr[1]=='N' )										// read out north our south hemisphere
					gps_north_south = 0;
				else
					gps_north_south = 1;
				
				substring_ptr = memchr(substring_ptr+1, ',', BUFFER_SIZE);		// get a pointer to the next position of a comma in the message
				
				lon_f = atof(substring_ptr+1);									// read the longitude as floating point value from the message
																				// (float looses some very slight precision, but is much more robust
																				//  against different numbers of digits in the GPS receiver implementation,
																				//  so this way should work with basically all kind of GPS receivers.
				
				gps_long_ddmm = (int)lon_f;										// extract the integer part of the longitude
				gps_long_dec_mmmm = (lon_f-gps_long_ddmm)*10000;				// extract the post-decimal part of the longitude as 4digit number as wanted by HoTT
		
				substring_ptr = memchr(substring_ptr+1, ',', BUFFER_SIZE);		// get a pointer to the next position of a comma in the message
		
				if( substring_ptr[1]=='W' )										// read out west or east 
					gps_east_west = 1;
				else
					gps_east_west = 0;
				
				substring_ptr = memchr(substring_ptr+1, ',', BUFFER_SIZE);		// get a pointer to the next position of a comma in the message
		
				speed_knots_f = atof(substring_ptr+1);							// read speed in knots from message
				gps_speed_kmh = (int)(speed_knots_f*1.852);						// convert knots to km/h and store as integer
				
				substring_ptr = memchr(substring_ptr+1, ',', BUFFER_SIZE);		// get a pointer to the next position of a comma in the message
				
				course_f = atof(substring_ptr+1);								// read course in degrees from message
				gps_flightdir_2deg = ((int)course_f)/2;
				
				// now we need data from the next data set, so we copy that one into the temporary buffer (again with interupts locked)
				cli();
				memcpy(temp_buffer, gps_buffer_gpgga, BUFFER_SIZE);
				sei();
				
				//example message from another GPS receiver as test vector
				//strcpy(temp_buffer, "$GPGGA,123519,4807.038,N,01131.000,E,1,08,0.9,545.4,M,46.9,M,,*47");
							
				substring_ptr = memchr(temp_buffer, ',', BUFFER_SIZE);			// get a pointer to the first position of a comma in the message
				// skip time info
				substring_ptr = memchr(substring_ptr+1, ',', BUFFER_SIZE);		// get a pointer to the next position of a comma in the message
				// skip latitude info
				substring_ptr = memchr(substring_ptr+1, ',', BUFFER_SIZE);		// get a pointer to the next position of a comma in the message
				// skip N/S info
				substring_ptr = memchr(substring_ptr+1, ',', BUFFER_SIZE);		// get a pointer to the next position of a comma in the message
				// skip longitude info
				substring_ptr = memchr(substring_ptr+1, ',', BUFFER_SIZE);		// get a pointer to the next position of a comma in the message
				// skip E/W info

				substring_ptr = memchr(substring_ptr+1, ',', BUFFER_SIZE);		// get a pointer to the next position of a comma in the message
				gps_fix_mode = atof(substring_ptr+1);							// read out the fix mode (DGPS, 2D, 3D)
				
				substring_ptr = memchr(substring_ptr+1, ',', BUFFER_SIZE);		// get a pointer to the next position of a comma in the message
				gps_no_sats = atof(substring_ptr+1);							// read out the number of sats used
				
				substring_ptr = memchr(substring_ptr+1, ',', BUFFER_SIZE);		// get a pointer to the next position of a comma in the message
				// skip HDOP info
				
				substring_ptr = memchr(substring_ptr+1, ',', BUFFER_SIZE);		// get a pointer to the next position of a comma in the message
				altitude_f = atof(substring_ptr+1);								// read out the altitude
				gps_altitude = (int)altitude_f+500; 							// HoTT protocol requires an offset of 500
				
				gps_distance = gps_time_hour*100 + gps_time_min; // use the home distance to show the gps time hours and minutes
				gps_dir_mod_home = gps_time_sec/2;	// we use the home direction to show the gps time seconds in the display (gets doubled in transmitter);

				
				// calculate climb/sink rate from the (rather unprecise in terms of altitude) GPS data
				if(second_tick>=3)												// a 3 second timer is derived from the 1 second message interval from the GPS
				{
					diff_altitude = gps_altitude - old_altitude;				// calculate the difference in altitude since last check, 3 seconds ago
					gps_climb_mp3s = 120 + diff_altitude ;						// HoTT protocol requires an offset of 120 for altitude
					old_altitude = gps_altitude;								// remember the altitude for the next check in 3 seconds
					second_tick=0;												// reset the timer
				} 
				
				
			}
		   
		   
			// --- all data was collected now, so lets transmit it to the HoTT remote control ---
			
			_delay_ms(3);                                   // protocol needs 5ms, we have 2ms more in send_byte_via_sw_uard(), so we wait 3ms here
			
			UART_OPMODE_TX;                                 // switch HoTT pin to output mode

			checksum = 0;                                   // reset checksum at start of HoTT telegram

			send_byte_via_sw_uart(HOTT_START_BYTE);			// 1 start byte constant
			send_byte_via_sw_uart(HOTT_GPS_MODULE_SIG);		// 2 sensor id 
			
			send_byte_via_sw_uart(gps_alarm_beep);			// 3 warning_beeps (triggers respective voice output), 0x00 = no alarm => tested ok 
															// 		1 = min speed, 2 = decent per 3s, 12 = max speed, 4 = max distance
															//		13 = ascent per 3s, 15 = min altitude, 26 = max altitude  		
			
			send_byte_via_sw_uart(HOTT_GPS_TEXT_ID);		// 4 sensor text id
			
			send_byte_via_sw_uart(gps_alarm_bitmask);		// 5 alarm bitmask, respective value gets red in display if bit is set => tested ok
															// 		1=distance, 2=speed, 4=height, 8=climb/descent m/s, 16=climb decent m/3s
			
			send_byte_via_sw_uart(gps_alarm_no_fix);		// 6 gps fix alarm, 0=ok, 1=no GPS signal, GPS coordinates text gets red in display if 1 => tested ok 
			
			send_byte_via_sw_uart(gps_flightdir_2deg);		// 7 Flight direction in steps of 2°, clockwise with 0°=N => tested ok
			
			send_byte_via_sw_uart(LSB(gps_speed_kmh));		// 8 Speed in km/h LSB (diplay in MZ-18 is limited to 3 decimal digits, wraps at 999) => tested ok
			send_byte_via_sw_uart(MSB(gps_speed_kmh));		// 9 Speed in km/h MSB 

			send_byte_via_sw_uart(gps_north_south);			// 10 Hemisphere North or South, north=0, south=1 => tested ok
			
			send_byte_via_sw_uart(LSB(gps_lat_ddmm));		// 11 Latitude as 4 digit number Degress and Minutes as DDMM, LSB => tested ok
			send_byte_via_sw_uart(MSB(gps_lat_ddmm));		// 12 Latitude as 4 digit number Degress and Minutes as DDMM, MSB

			send_byte_via_sw_uart(LSB(gps_lat_dec_mmmm));	// 13 Latitude decimal, 4 digits, LSB (simply shown after decimal point) => tested ok 
			send_byte_via_sw_uart(MSB(gps_lat_dec_mmmm));	// 14 Latitude decimal, 4 digits, MSB 

			send_byte_via_sw_uart(gps_east_west);			// 15 Worlds East or West, east=0, west=1 => tested ok

			send_byte_via_sw_uart(LSB(gps_long_ddmm));		// 16 Longitude as 4 digit number Degress and Minutes as DDMM, LSB => tested ok
			send_byte_via_sw_uart(MSB(gps_long_ddmm));		// 17 Longitude as 4 digit number Degress and Minutes as DDMM, MSB
			
			send_byte_via_sw_uart(LSB(gps_long_dec_mmmm));	// 18 Longitude decimal, 4 digits, LSB (simply shown after decimal point) => tested ok 
			send_byte_via_sw_uart(MSB(gps_long_dec_mmmm));	// 19 Longitude decimal, 4 digits, MSB => tested ok
			
			send_byte_via_sw_uart(LSB(gps_distance));		// 20 Distance from model to home in meters LSB (works to full number range of 0xffff) => tested ok 
			send_byte_via_sw_uart(MSB(gps_distance));		// 21 Distance from model to home in meters MSB 
			
			send_byte_via_sw_uart(LSB(gps_altitude)); 		// 22 Altitude in steps of 1m, with -500m offset, i.e. 500=0m, LSB => tested ok
			send_byte_via_sw_uart(MSB(gps_altitude)); 		// 23 Altitude in steps of 1m, with -500m offset, i.e. 500=0m, MSB 
			
			send_byte_via_sw_uart(LSB(gps_climb_mps));		// 24 Climb rate for m/s gauge in steps of 0.01m/s, with -30000 offset, i.e. 30000 = 0m/s LSB => tested ok
			send_byte_via_sw_uart(MSB(gps_climb_mps));		// 25 Climb rate for m/s gauge in steps of 0.01m/s, with -30000 offset, i.e. 30000 = 0m/s MSB
			
			send_byte_via_sw_uart(gps_climb_mp3s);			// 26 Climb rate for m/3s gauge with -120 offset, i.e. 120 = 0m/3s => tested ok
			
			send_byte_via_sw_uart(gps_no_sats);				// 27 Nnumber of satelites => tested ok
			
			send_byte_via_sw_uart(gps_fix_mode);			// 28 GPS fix character, display if DGPS, 2D oder 3D => apparently not shown in display
			
			send_byte_via_sw_uart(gps_dir_mod_home);		// 29 Direction from model to home in steps of 2° => tested ok
			
			send_byte_via_sw_uart(0);						// 30 angle x-direction (1 byte) => apparently not shown in display
			send_byte_via_sw_uart(0);						// 31 angle y-direction (1 byte) => apparently not shown in display
			send_byte_via_sw_uart(0);						// 32 angle z-direction (1 byte) => apparently not shown in display
			
			send_byte_via_sw_uart(gps_time_hour);			// 33 UTC time hours / gyro x low byte (2 bytes) => apparently not shown in display
			send_byte_via_sw_uart(gps_time_min);			// 34 UTC time minutes / gyro x high byte => apparently not shown in display
			send_byte_via_sw_uart(gps_time_sec);			// 35 UTC time seconds / gyro y low byte (2 bytes) => apparently not shown in display
			send_byte_via_sw_uart(0);						// 36 UTC time milliseconds / gyro y high byt e=> apparently not shown in display
			
			send_byte_via_sw_uart(0);						// 37 mean sea level altitude / gyro z low byte (2 bytes) => apparently not shown in display
			send_byte_via_sw_uart(0);						// 38 gyro z high byte => apparently not shown in display
			
			send_byte_via_sw_uart(0);						// 39 vibration (1 bytes) => apparently not shown in display
			
			send_byte_via_sw_uart('a');						// 40 00 ASCII Free Character [4] => apparently not shown in display
			send_byte_via_sw_uart('b');						// 41 00 ASCII Free Character [5] => apparently not shown in display
			send_byte_via_sw_uart('c');						// 42 00 ASCII Free Character [6] => apparently not shown in display
			
			send_byte_via_sw_uart(0);						// 43 Version number => apparently not shown in display

			send_byte_via_sw_uart(0x7d);					// 44 stop byte 0x7d
			send_byte_via_sw_uart(checksum & 0xFF);			// 45 checksum (last 8bits of sum of all data bytes)
			
			_delay_ms(1);                                   // pause of 1ms before we switch back to receive mode
			UART_OPMODE_RX;                                 // switch back to receive mode
	
	
		} // end of: if( received_databyte == HOTT_GPS_MODULE_SIG )	
    
		
		if( received_databyte == HOTT_GEN_MODULE_SIG )		// if 2nd byte of the request is that of the general sensor...  
		{
		
			// --- now we know we have been called and shall respond with our information, lets put that together ---
    	
			_delay_ms(3);                                   // protocol needs 5ms, we have 2ms more in send_byte_via_sw_uard(), so we wait 3ms here
			
			UART_OPMODE_TX;                                 // switch HoTT pin to output mode

			checksum = 0;                                   // reset checksum at start of HoTT telegram

			send_byte_via_sw_uart(HOTT_START_BYTE);			// 1 start byte constant (always 0x7c)
			send_byte_via_sw_uart(HOTT_GEN_MODULE_SIG);		// 2 sensor id (0x8d = general module)
			send_byte_via_sw_uart(alarm_beep);				// 3 warning_beeps (triggers respective voice output), 0x10= low voltage, 0x00 = no alarm => tested ok
			send_byte_via_sw_uart(0xd0);					// 4 constant value 0xd0
			send_byte_via_sw_uart(0x00);					// 5 alarm bitmask inverted, 0 all cell voltage, 1 Battery 1, 2 Battery 2, 3 Temperature 1, 4 Temperature 2, 5 Fuel, 6 mAh, 7 Altitude
			send_byte_via_sw_uart(0x00);					// 6 alarm bitmask inverted, 0 main power current, 1 main power voltage, 2 Altitude, 3 m/s, 4 m/3s// 5 unknown// 6 unknown// 7 "ON" sign/text msg active
			send_byte_via_sw_uart(cell1_voltage);			// 7 voltage cell 1 in 20mV => tested ok
			send_byte_via_sw_uart(cell2_voltage);			// 8 voltage cell 2 in 20mV => tested ok
			send_byte_via_sw_uart(cell3_voltage);			// 9 voltage cell 3 in 20mV => tested ok
			send_byte_via_sw_uart(0x00);					// 10 voltage cell 4 in 20mV => tested ok, hard coded to 0 here since we support only 3S battery
			send_byte_via_sw_uart(0x00);					// 11 voltage cell 5 in 20mV => tested ok, hard coded to 0 here since we support only 3S battery
			send_byte_via_sw_uart(0x00);					// 12 voltage cell 6 in 20mV => tested ok, hard coded to 0 here since we support only 3S battery
			send_byte_via_sw_uart(battery_voltage_in_100mV);// 13 overall voltage battery 1 in 100mV (LSB) => tested ok
			send_byte_via_sw_uart(0x00);					// 14 overall voltage battery 1 in 100mV (MSB) => tested ok (set to 0x00 here, as a 3S battery does not reach 25.6V)
			send_byte_via_sw_uart(0x00);					// 15 overall voltage battery 2 in 100mV (LSB)  => tested ok
			send_byte_via_sw_uart(0x00);					// 16 overall voltage battery 2 in 100mV (MSB)   => tested ok
			send_byte_via_sw_uart(temperature1);			// 17 temperature 1 (offset 20, 20=0°C) => tested ok
			send_byte_via_sw_uart(temperature2);			// 18 temperature 2 (offset 20, 20=0°C) => tested ok
			send_byte_via_sw_uart(fuel_bar);				// 19 fuel shown as % in bargraph (0..100) => tested ok
			send_byte_via_sw_uart(LSB(battery_voltage_in_mV));// 20 fuel shown in ml as text (full=65535) LSB, we use it for voltage in mV => tested ok (sidenote: value stalls at 65000) 
			send_byte_via_sw_uart(MSB(battery_voltage_in_mV));// 21 fuel shown in ml as text (full=65535) MSB => tested ok (see line above)
			send_byte_via_sw_uart(0x00);    				// 22 rpm in steps of 10 (100=1000rpm) LSB => tested ok
			send_byte_via_sw_uart(0x00);					// 23 rpm in steps of 10 (100=1000rpm) MSB => tested ok
			send_byte_via_sw_uart(0xf4);					// 24 altitude in m (500 offset, 0x01f4=500=0m) LSB => tested ok (result only shown as text)
			send_byte_via_sw_uart(0x01);					// 25 altitude in m (500 offset, 0x01fa=500=0m) MSB => tested ok (result only shown as text)
			send_byte_via_sw_uart(0x30);					// 26 climb rate in 0.01ms/s, 0x7530=30000=0.00m/s LSB => tested ok
			send_byte_via_sw_uart(0x75);					// 27 climb rate in 0.01ms/s, 0x7530=30000=0.00m/s MSB => tested ok
			send_byte_via_sw_uart(0x78);                    // 28 climb rate in m/3s, 0x78=120=0m/3s => tested ok
			send_byte_via_sw_uart(0x00);					// 29 current in 100mA, 100=10A LSB => tested ok
			send_byte_via_sw_uart(0x00);					// 30 current in 100mA, 100=10A MSB => tested ok
			send_byte_via_sw_uart(battery_voltage_in_100mV);// 31 main power voltage in 100mV (shown in overview screen) LSB => tested ok
			send_byte_via_sw_uart(0x00);					// 32 main power voltage in 100mV (shown in overview screen) MSB => tested ok (set to 0x00 here, as a 3S battery does not reach 25.6V)
			send_byte_via_sw_uart(0x00);					// 33 used battery capacity in 10mAh, LSB => tested ok
			send_byte_via_sw_uart(0x00);					// 34 used battery capacity in 10mAh, MSB => tested ok
			send_byte_via_sw_uart(0x00);					// 35 speed LSB => not tested
			send_byte_via_sw_uart(0x00);					// 36 speed MSB => not tested
			send_byte_via_sw_uart(0x00);					// 37 minimum cell voltage in 20mV steps? => not tested
			send_byte_via_sw_uart(0x00);					// 38 number of the cell with the lowest voltage => not tested
			send_byte_via_sw_uart(0x00);					// 39 rpm 2 in steps of 10 (100=1000rpm) LSB => not tested
			send_byte_via_sw_uart(0x00);					// 40 rpm 2 in steps of 10 (100=1000rpm) MSB => not tested
			send_byte_via_sw_uart(0x00);					// 41 general error number (12=voice) => not tested
			send_byte_via_sw_uart(0x00);					// 42 high pressure up to 16bar in 100mbar  => not tested
			send_byte_via_sw_uart(0x00);					// 43 version number => not tested
			send_byte_via_sw_uart(0x7d);					// 44 stop byte 0x7d
			send_byte_via_sw_uart(checksum & 0xFF);			// 45 checksum (last 8bits of sum of all data bytes)
			
			_delay_ms(1);                                   // pause of 1ms before we switch back to receive mode
			UART_OPMODE_RX;                                 // switch back to receive mode
			
			// --- now we can take care to get new measurement values from the ADCs and fill our status variable with the latest data ---

			read_cell_voltages();							// trigger the ADC reading, returns the cell voltages in 20mV
															// and the overall battery voltage in mV and in 100mV, and fuel bar in %

			// if one of the cell voltages gets too low...
			if(cell1_voltage <= LOW_CELL_VOLTAGE_THRESHOLD || cell2_voltage <= LOW_CELL_VOLTAGE_THRESHOLD || cell3_voltage <= LOW_CELL_VOLTAGE_THRESHOLD)
				alarm_beep = 0x10;							// trigger acoustic warning at remote control
			else
				alarm_beep = 0x00;							// no warning when voltage is ok on all cells
		
    
		} // end of: if( received_databyte == HOTT_GPS_MODULE_SIG )
    
    }

    return 0;
    
}


// --------------------------------- reset handler -----------------------------
void reset_handler(void)
{
   
	init_RS232_timer();				    // start a clock counter for 19k2 baudrate
     
	LED_DDR |= LED_BIT;				    // set LED pin as output
    LED_OFF;                            // start with LED off
	
	UART_OPMODE_RX;					    // the HoTT interface pin to high impedance first (receiving)
    UART_TXD_HIGH;					    // keep the logical value on the HoTT high from the start, otherwise
                                        //   switching to TX mode will cause a falling edge right away which
                                        //   would erronously be detected as 0x00 data byte the the HoTT receiver

	UBRRL = 71;							// Configure hardware UART to receive data from GPS as 9k6 8N1 (fc=11.059MHz)
	UCSRB |= _BV(RXEN) | _BV(TXEN) | _BV(RXCIE);	// Enable UART RX operation, UART TX operation (for debugging) and 
													// enable the interupt generation at UART RX complete

    sei();                              // Global interupt enable

}


// ------------------------------- send byte via uart --------------------------
void send_byte_via_sw_uart(char byte_to_send)
{
    uint8_t bit;    
    
    _delay_ms(2);						// the Hott protocol wants delays between data bytes,
										//   we place this delay generically here, rather than
										//	 for every call of this function, to save memory
    
    checksum += byte_to_send;           // to add up the checksum for the complete protocol here
                                        //   is a way to safe memory
    
    wait_for_next_uart_cycle();         // synchronize with the next overflow of the timer
    
    UART_TXD_LOW;                       // send startbit (always 0)
    
 
    for( bit=0 ; bit<=7 ; bit++ )       // send all 8 bits of this byte, starting with LSB
    {
       
        wait_for_next_uart_cycle();     // synchronize with the next overflow of the timer
        
        if(byte_to_send & _BV(0))       // take TXD line to high if LSB is set
        {
            UART_TXD_HIGH;
        }
        else
        {
            UART_TXD_LOW; 				// respectively take it low if LSB is not set
        }
        
        byte_to_send>>=1;               // shift the byte to be send by one bit towards LSB
    
    }   
    
    wait_for_next_uart_cycle();         // synchronize with the next overflow of the timer

    UART_TXD_HIGH;                      // send stopbit (always 1)
    
}


// ------------------------------ receive byte via uart (blocking)------------------------
int receive_byte_via_sw_uart(void)
{
  
    unsigned char bit_counter;
    unsigned char byte_received=0;
  
	UART_OPMODE_RX;								// switch the joint UART pin to RX mode
    
    while( UART_RXD );                          // wait for UART RXD to go low (beginning of startbit)
												// (this is the most frequented waiting loop)
    
    TCNT1 = 0;                                  // reset counter to synchronize on incoming bits
	
                                                // read 8 data bit (plus start bit)
    for( bit_counter = 0; bit_counter < 9 ; bit_counter++ )
    {

        byte_received >>= 1;                    // shift the already received bits one right
            
        TIFR |= _BV(OCF1B);                    	// clear output compare B match flag
        while( !(TIFR & _BV(OCF1B)) );         	// wait for middle of bit
          
        if( UART_RXD )                          // sample on RXD, if it is high...
        {
            byte_received |= _BV(7);            // ...set MSB in received byte
        }
        
    }

    // now only the stop bit is left, which we simply ignore, as it is always high
    TIFR |= _BV(OCF1B);                    		// clear output compare B match flag
    while( !(TIFR & _BV(OCF1B)) );         		// wait for middle of bit

    // intentionally we do not wait for the end of the stop bit here, that gives us
    // half a bit of time to process the received byte, before the next one may arive
    // (waiting for end of stop bit could be done with: wait_for_next_uart_cycle(); )

    return byte_received;
    
}


// ----------- synchronize with the next overflow of timer1 -----------------
void wait_for_next_uart_cycle(void)
{
    
    TIFR |= _BV(OCF1A);                            // clear the timer overflow flag
    while( !( TIFR & _BV(OCF1A)) );                // wait until overflow flag is set again

}


// --- switch on a permanent clock source that runs at the baudrate of 19200kHz
void init_RS232_timer(void)
{
 
    // baudrate is 19200 bit/s => bit length is 52.08µs
    // main clock is 11.059MHz, with prescaler 1 => clock period 90.42ns
    // => 576 clock cycles per bit length
 
    // set counter compare match to 576-1 = 575
    // (the -1 is due to the fact that the value is actually reached before the counter is reset)
    OCR1A = 575;        	// compiler can handle the 16bit register access, simply assign the number
    
    // for sampling incoming bits in the middle, use the second compare match unit at half that value
    OCR1B = 286;
 
	TCCR1B |= _BV(WGM12); 	// CTC mode =  clear time on compare match, timer wraps to 0 when OCR1A value reached

	TCCR1B |= _BV(CS10); 	// Switch timer on, to run without prescaler, i.e. at full system clock speed (3.686MHz)

    TCNT1 = 0;              // Reset the counter, so it starts to count at zero now
    
}

// extract the upper byte from a 16bit word
uint8_t MSB(uint16_t word_var)
{
	
	return (word_var >> 8);
	
}

// extract the lower byte from a 16bit word
uint8_t LSB(uint16_t word_var)
{
	
	return (word_var & 0xff);
	
}

// UART receive complete interupt service routine (called every time a new byte came in from GPS)
// This routine is a bit strangely made, that is to safe as much time as possible. Any runtime beyond
// 10-15µs would affect the ongoing serial communication on the HoTT interface too much.
ISR(USART_RXC_vect)
{

	static uint8_t recording_gprmc_on;
	static uint8_t recording_gpgga_on;
	static uint8_t buffer_pos=0;
	uint8_t received_gps_byte;
    
    received_gps_byte = UDR;								// store the received character
    
    if(received_gps_byte=='$')								// $ means start of new msg, we assume for now it is the right one (GPRMC)
	{
		buffer_pos = 0;										// reset the postion counter for the buffer
		recording_gprmc_on = 1;								// start recording of incoming data for msg GPRMC
		recording_gpgga_on = 1;								// start recording of incoming data for msg GPGGA
	}
	
        
    if(recording_gprmc_on)
    {
		if(buffer_pos==5)
		{
			if(received_gps_byte!='C')						// if 5th character after $ is not C, this is not the right msg
			{
				recording_gprmc_on = 0;						// so we terminate recording, to avoid earlier (valid) data to be overwritten
			}
			else
			{
				second_tick++;								// the message comes every 1 second -> we use it at a rough timer
			}
		}
	
		gps_buffer_gprmc[buffer_pos] = received_gps_byte;	// store the received character
	}

    if(recording_gpgga_on)
    {
		if(buffer_pos==4 && (received_gps_byte!='G'))		// if 4th character after $ is not G, this is not the right msg
			recording_gpgga_on = 0;							// so we terminate recording, to avoid earlier (valid) data to be overwritten
	
		gps_buffer_gpgga[buffer_pos] = received_gps_byte;	// store the received character
	}	
	
	if(buffer_pos<(BUFFER_SIZE-1))							// if buffer is not full yet
		buffer_pos++;										// increment the postion counter

}

//------------------------------- read adc -----------------------
unsigned int read_adc(unsigned char adc_number)
{

    uint16_t adc_result;
    
    ADCSRA |= _BV(ADPS2) | _BV(ADPS1) | _BV(ADPS0); // ADC prescaler :128, that 
                                        // gives ADC frequency of 11.059MHz/128 = 86kHz

    ADMUX = adc_number;                 // select Vcc reference (Bits 7 and 6 remain zero)
                                        // and chose the ADC channel

    ADCSRA |= _BV(ADEN);                // switch on the ADC in general

    ADCSRA |= _BV(ADSC);                // start a single ADC conversion
    while( ADCSRA & _BV(ADSC) );        // wait until conversion is complete

    // data sheet recommends to discard first conversion, so we do one more
    ADCSRA |= _BV(ADSC);                // start a single ADC conversion
    while( ADCSRA & _BV(ADSC) );        // wait until conversion is complete

    adc_result = ADC;                   // take over ADC reading result into variable

    ADCSRA = 0;                         // completely disable the ADC to save power
    	
	return adc_result;
	
}


// ----- read cell voltages via a 1:4.3 voltage divider each, reference is 3.3V -----
// returns cell voltages in 20mV steps, total battery voltage in 100mV and the fuel bar indicator in percent
void read_cell_voltages(void)
{
	
	// For simplified hardware, this function reads only one ADC (the overall battery voltage) and splits up
	// the voltage equaly over all 3 cells. 

	uint16_t adc_result;
	uint16_t avg_sum=0;
	uint8_t a;
	uint16_t cell_voltage_in_20mV;
	
	// Correction factor for slightly too low voltage at ADC input, caused by rather big resitstors in the 
	// 1:4.3 voltage divider (10K/33K). The measured error is 0.8%, that needs to be compensated for.
	#define CORR_FACTOR						1.008
	
	// Factor for calculating single cell voltages and for the overall battery voltage is derived like this: 
	// With 1:4.3 resistive divider (10K/33K) and 3.3V reference, the max ADC value corresponds to 14.19V.
	// At 10bit ADC resolution, one ADC step is 13.85mV on the battery. 
	#define ADC_CALC_FACTOR_BATT_IN_MV 		(13.85*CORR_FACTOR)
	// The single cell voltage is in 20mV and is 1/3 of the overall voltage, i.e. 0,01385 / 0.020 / 3 = 0.2308 per ADC step.
	#define ADC_CALC_FACTOR_SINGLE_CELL 	(0.2308*CORR_FACTOR)
	// The overall battery voltage is in 100mV, i.e. 0.01385 / 0.1 = 0.1385 per ADC step.
	#define ADC_CALC_FACTOR_BATT_VLT 		(0.1385*CORR_FACTOR)
	
	
	for(a=0; a<8; a++)                              	// do an averaging over 8 ADC reads for each cell
    {
        adc_result = read_adc(LIPO_TAP3_ADC_CHANNEL);	// read ADC for complete battery voltage (tap3 for a 3S battery)
        avg_sum += adc_result;                      	// add up the results on cell1 for the averaging
        
        _delay_ms(1);                               	// wait 1ms for good independence between single reads
    }
	
	avg_sum >>= 3;                                  	// divide the added results by 8 for final result
	
	battery_voltage_in_mV = (ADC_CALC_FACTOR_BATT_IN_MV * avg_sum)+0.5;		// calculate in floating point, then round correctly to integer
																			// and write the result to the return value variable directly

	cell_voltage_in_20mV = (ADC_CALC_FACTOR_SINGLE_CELL * avg_sum)+0.5;		// calculate in floating point, then round correctly to integer
	cell1_voltage = cell_voltage_in_20mV;									// set all single cell voltages to the same calculated value
	cell2_voltage = cell_voltage_in_20mV;
	cell3_voltage = cell_voltage_in_20mV;
		
	battery_voltage_in_100mV =  (ADC_CALC_FACTOR_BATT_VLT * avg_sum)+0.5;	// calculate in floating point, then round correctly to integer
																			// and write the result to the return value variable directly
    
    // calculate the fuel bar gauge reading from the measured battery voltage and the defined upper and lower limit for the gauge
    float fuel_bar_beyond_min = (float)(battery_voltage_in_mV) - (float)(LOW_FUEL_BAR_VALUE);	// first calculate the voltage above minimum
    if( fuel_bar_beyond_min > 0)																// if we are above the minimum
    {
		fuel_bar = (uint8_t)( (fuel_bar_beyond_min*100.0) / (float)(MAX_FUEL_BAR_VALUE - LOW_FUEL_BAR_VALUE) );	// calculate factor in %
    }
    else 																						// if we are below the minimum
    {
		fuel_bar = 0;																			// we simply show zero, to avoid an underrun
	}
    
    	    	
}


